# -*- coding: utf-8 -*-


class OlsException(Exception):
    """
    A base class for all `OLS` exceptions.
    """
    def __init__(self, error):
        self.error = error['message']
        self.status = error['status']
        self.origin = error['exception']
        self.timestamp = error['timestamp']

    def __repr__(self):
        return '{}({}) [{}-{}]'.format(self.origin, self.status, self.timestamp, self.error)

    def __str__(self):
        return str(self.error)


class NotFoundException(OlsException):
    """
    Entity was not found via a OLS api call
    """
    pass


class BadParameter(OlsException):
    """
    Passed parameters does not fit with expected
    """
    pass




# -*- coding: utf-8 -*-

from coreapi import Client

import ebi.ols.api.helpers as helpers
from ebi.ols.api.base import decoders, site, DetailClientMixin, ListClientMixin, SearchClientMixin


class OlsClient(object):
    """
    TODO: add individuals and properties services
    OLS Client ontologies client

    """
    class DetailClient(object):

        def __call__(self, item, **kwargs):
            if isinstance(item, helpers.Property):
                client = DetailClientMixin('ontologies/{}/properties'.format(item.ontology_name), helpers.Property)
                return client(item.iri)
            elif isinstance(item, helpers.Individual):
                client = DetailClientMixin('ontologies/{}/individuals'.format(item.ontology_name), helpers.Ontology)
            elif isinstance(item, helpers.Ontology):
                client = DetailClientMixin('ontologies', helpers.Ontology)
            elif isinstance(item, helpers.Term):
                client = DetailClientMixin('ontologies/{}/terms'.format(item.ontology_name), helpers.Term)
            else:
                assert('ontology_name' in kwargs)
                assert('iri' in kwargs)
                assert(issubclass(item, helpers.OLSHelper))
                inst = item(ontology_name=kwargs.get('ontology_name'), iri=kwargs.get('iri'))
                return self.__call__(inst)
            return client(item.iri)

    def __init__(self):
        self.client = Client(decoders=decoders)
        document = self.client.get(site)
        self.ontologies = ListClientMixin('ontologies', helpers.Ontology, document)
        self.ontology = DetailClientMixin('ontologies', helpers.Ontology)
        self.terms = ListClientMixin('terms', helpers.Term, document)
        self.term = DetailClientMixin('terms', helpers.Term)
        self.search = SearchClientMixin('search', helpers.Term, document)
        self.detail = self.DetailClient()


if __name__ == "__main__":
    # TODO add arg parse and associated method calls
    print('Coming soon')



# -*- coding: utf-8 -*-

import re
from collections import namedtuple, OrderedDict

import inflection
from coreapi import Object

from ebi.ols.api.base import *


def underscore(value):
    val = inflection.underscore(value)
    return re.sub('\s+', '_', val)


def convert_keys(data):
    """
    Convert Json like keyx into Python-like keys
    :param data: a OrderedDict or None
    :return: OrderedDict
    """
    if data is None:
        return OrderedDict({})
    return OrderedDict(
        {underscore(k): convert_keys(v) if isinstance(v, dict) or isinstance(v, Object) else v for k, v in
         data.items()})


def to_python_value(value):
    type_value = type(value)
    if type_value is bool:
        return value
    elif type_value is 'str':
        if value in ('true', 'True', '1'):
            return True
        elif value in ('0', 'false', 'False'):
            return False
    return value


class OLSHelper(object):
    """
    Base Transfer object, mainly assign dynamically received dict keys to object attributes
    """

    def __init__(self, **kwargs):
        converted = convert_keys(kwargs)
        for name, value in converted.items():
            # print(name, value, to_python_value(value))
            self.__setattr__(name, to_python_value(value))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)


class OntologyAnnotation(OLSHelper):
    """
    An ontology annotation item.
    """
    license = None
    creator = []
    rights = []
    format_version = None
    comment = []
    default_namespace = None
    contributor = []


class OntologyConfig(OLSHelper):
    id = None
    version_iri = None
    title = None
    namespace = None
    preferred_prefix = None
    description = None
    homepage = None
    version = None
    mailing_list = None
    creators = []
    annotations = {}
    file_location = None
    reasoner_type = None
    obo_slims = None
    label_property = None
    definition_properties = []
    synonym_properties = []
    hierarchical_properties = []
    base_uris = []
    hidden_properties = []
    internal_metadata_properties = []
    skos = None

    def __init__(self, **kwargs):
        annotations = OntologyAnnotation(**kwargs.pop("annotations", {}))
        super().__init__(annotations=annotations, **kwargs)

    def __repr__(self):
        return '<OntologyConfig(id={}, iri={}, namespace={}, title={}, version={})>'.format(
            self.id, self.version_iri, self.namespace, self.title, self.version)


class Ontology(OLSHelper):
    """
    An Ontology element from EBI Ontology lookup service
    """
    path = 'ontologies'

    ontology_id = None
    loaded = None
    updated = None
    status = None
    message = None
    _version = None
    number_of_terms = None
    number_of_properties = None
    number_of_individuals = None
    config = None
    annotations = {}

    def __init__(self, **kwargs):
        config = OntologyConfig(**kwargs.pop("config", {}))
        super().__init__(config=config, **kwargs)

    def __repr__(self):
        return '<Ontology(ontology_id={}, title={}, namespace={}, updated={})>'.format(
            self.ontology_id, self.config.title, self.config.namespace, self.updated)

    def terms(self, filters=None):
        """ Links to ontology associated terms"""
        client = ListClientMixin('ontologies/' + self.ontology_id, Term)
        return client(filters=filters)

    def individuals(self, filters=None):
        """ Links to ontology associated individuals """
        client = ListClientMixin('ontologies/' + self.ontology_id, Individual)
        return client(filters=filters)

    def properties(self, filters=None):
        """ Links to ontology associated properties"""
        client = ListClientMixin('ontologies/' + self.ontology_id, Property)
        return client(filters=filters)

    @property
    def namespace(self):
        if self.config.annotations.default_namespace:
            return self.config.annotations.default_namespace[0]
        else:
            return self.config.namespace

    @property
    def title(self):
        return self.config.title

    @property
    def version(self):
        return self._version if self._version else self.config.version

    @version.setter
    def version(self, version):
        self._version = version

    @namespace.setter
    def namespace(self, namespace):
        if self.config.annotations.default_namespace:
            self.config.annotations.default_namespace[0] = namespace
        else:
            self.config.namespace = namespace


class OboXref(OLSHelper):
    database = None
    id = None
    description = None
    ur = None


class OboCitation(OLSHelper):
    definition = None
    oboXref = None


class TermAnnotation(OLSHelper):
    database_cross_reference = None
    has_obo_namespace = []
    has_alternative_id = []
    id = []
    alternative_term = []
    definition_source = []
    editor_preferred_term = []
    example_of_usage = []
    term_editor = []
    term_tracker_item = []

    def __repr__(self):
        return '<Term(id={}, has_obo_name_space={}, alternative_term={}, alt_id={})>'.format(
            self.id, self.has_obo_namespace, self.alternative_term, self.has_alternative_id)


class TermsLink(OLSHelper):
    parents = None
    ancestors = None
    hierarchicalParents = None
    hierarchicalAncestors = None
    jstree = None
    children = None
    descendants = None
    hierarchicalChildren = None
    hierarchicalDescendants = None
    graph = None


class Term(OLSHelper):
    path = 'terms'
    iri = None
    label = None
    _description = []
    annotation = None
    synonyms = None
    ontology_name = None
    ontology_prefix = None
    ontology_iri = None
    is_obsolete = False
    term_replaced_by = None
    is_defining_ontology = False
    has_children = False
    is_root = False
    short_form = None
    obo_id = None
    in_subset = None
    obo_definition_citation = None
    obo_xref = None
    obo_synonym = None

    def __init__(self, **kwargs):
        annotation = TermAnnotation(**kwargs.pop("annotation", {}))
        super().__init__(annotation=annotation, **kwargs)
        self._relations_types = None

    def __repr__(self):
        return '<Term(obo_id={}, name={}, ontology_id={}, namespace={} subsets={}, short_form={})>'.format(
            self.obo_id, self.label, self.ontology_name, self.obo_name_space, self.in_subset, self.short_form)

    @property
    def relations_types(self):
        if self._relations_types is None:
            client = ListClientMixin('ontologies/' + self.ontology_name + '/terms/' + uri_terms(self.iri), Term)
            self._relations_types = [name for name in convert_keys(client.document.links).keys() if
                                     name not in ('graph', 'jstree', 'descendants', 'ancestors', 'hierarchical_parents',
                                                  'hierarchical_ancestors', 'hierarchical_children',
                                                  'hierarchical_descendants')]
        # print(self._relations_types)
        return self._relations_types

    def load_relation(self, relation):
        if relation in self.relations_types:
            client = ListClientMixin('ontologies/' + self.ontology_name + '/terms/' + uri_terms(self.iri), Term)
            return client(action=relation)
        return []

    def children(self):
        return self.load_relation('children') if self.has_children else []

    def descendants(self):
        return self.load_relation('descendants')

    def ancestors(self):
        return self.load_relation('ancestors')

    def parents(self):
        return self.load_relation('parents')

    def hierarchical_parents(self):
        return self.load_relation('hierarchicalParents')

    def hierarchical_ancestors(self):
        return self.load_relation('hierarchicalAncestors')

    def hierarchical_children(self):
        return self.load_relation('hierarchicalChildren')

    def hierarchical_descendants(self):
        return self.load_relation('hierarchicalDescendants')

    def graph(self):
        return self.load_relation('graph')

    def jstree(self):
        return self.load_relation('jstree')

    @property
    def obo_name_space(self):
        # print(self.annotation)
        if self.annotation.has_obo_namespace:
            # print('has obo namespace', self.annotation.has_obo_namespace[0])
            return self.annotation.has_obo_namespace[0]
        else:
            return self.ontology_name

    @property
    def description(self):
        return self._description[0] if len(self._description) > 0 else ''

    @description.setter
    def description(self, value):
        if value is None:
            value = [self.label]
        self._description = value

    @property
    def subsets(self):
        return ','.join(
            [subset for subset in sorted(self.in_subset, key=lambda s: s.lower())]) if self.in_subset else ''

    @property
    def accession(self):
        return self.obo_id

    @property
    def name(self):
        return self.label


Subset = namedtuple("Subset", ["terms"])


class Individual(OLSHelper):
    path = 'individuals'
    iri = None
    label = None
    description = []
    synonyms = None
    ontology_name = None
    ontology_prefix = None
    ontology_iri = None
    is_obsoloete = None
    is_defining_ontology = None
    short_form = None
    obo_id = None

    def __repr__(self):
        return '<Individual(label={}, iri={}, ontology_name={}, short_form={}, obo_id={})>'.format(
            self.label, self.ontology_name, self.iri, self.short_form, self.obo_id)


class Property(OLSHelper):
    path = 'properties'
    annotation = None
    synonyms = None
    iri = None
    label = None
    description = None
    ontology_name = None
    ontology_prefix = None
    ontology_iri = None
    is_obsolete = None
    is_defining_ontology = None
    has_children = None
    is_root = None
    short_form = None
    obo_id = None

    def __repr__(self):
        return '<Property(label={}, iri={}, ontology_name={}, short_form={}, obo_id={})>'.format(
            self.label, self.ontology_name, self.iri, self.short_form, self.obo_id)

    @property
    def definition(self):
        return self.description


# -*- coding: utf-8 -*-
import math
import sys
import time
import urllib.parse
import warnings

from coreapi import codecs, Client
from coreapi.exceptions import ErrorMessage

from ebi.ols.api import exceptions
from ebi.ols.api.codec import HALCodec

decoders = [HALCodec(), codecs.JSONCodec()]
site = 'https://www.ebi.ac.uk/ols/api'


def filters_ontologies(filters):
    return filters or {}


def filters_response(filters):
    assert set(filters.keys()).issubset(
        {'ontology', 'type', 'slim', 'queryFields', 'exact', 'fieldList', 'groupField', 'obsoletes',
         'local'}) or len(filters) == 0, "Unauthorized filter key"
    if 'fieldList' in filters:
        assert (set(filters['fieldList'].keys().issubset(
            {'iri', 'ontology_name', 'ontology_prefix', 'short_form', 'description', 'id', 'label',
             'is_defining_ontology', 'obo_id', 'type'}
        )))
    if 'queryFields' in filters:
        assert (set(filters['queryFields'].keys().issubset(
            {'label', 'synonym', 'description', 'short_form', 'obo_id', 'annotations', 'logical_description',
             'iri'}
        )))
    if 'type' in filters:
        assert (set(filters['type'].keys().issubset(
            {'class', 'property', 'individual', 'ontology'}
        )))
    if 'exact' in filters:
        assert (filters['exact'] in ['true', 'false'])
    if 'groupFields' in filters:
        assert (filters['groupFields'] in ['true', 'false'])
    if 'obsoletes' in filters:
        assert (filters['obsoletes'] in ['true', 'false'])
    if 'local' in filters:
        assert (filters['local'] in ['true', 'false'])
    return filters


def filters_terms(filters):
    assert set(filters.keys()).issubset({'iri', 'obo_id', 'short_form'}), "Unauthorized filter key"
    assert len(filters.keys()) <= 1, "Only one filter can be applied at a time"
    return filters


def uri_ontologies(identifier):
    return identifier


def uri_terms(identifier):
    return urllib.parse.quote_plus(urllib.parse.quote_plus(identifier))


def uri_individuals(identifier):
    return urllib.parse.quote_plus(urllib.parse.quote_plus(identifier))


def uri_properties(identifier):
    return urllib.parse.quote_plus(urllib.parse.quote_plus(identifier))


class BaseClient(object):

    def __init__(self):
        self.client = Client(decoders=decoders)


class DetailClientMixin(BaseClient):

    def __init__(self, uri, elem_class):
        super().__init__()
        self.uri = uri
        self.elem_class = elem_class

    def __call__(self, identifier, silent=False, unique=False):
        """ Check one element from OLS API accroding to specified identifier
        In cas API returns multiple element return either:
        - the one which is defining_ontology (flag True)
        - The first one if none (Should not happen)
        """
        identifier_fn = getattr(sys.modules[__name__], 'uri_' + self.elem_class.path)
        path = "/".join([site, self.uri, identifier_fn(identifier)])
        try:
            document = self.client.get(path)
            # print('current path', self.elem_class, path)
            if self.elem_class.path in document.data:
                # the request returned a list of object
                if not silent:
                    warnings.warn('OLS returned multiple {}s for {}'.format(self.elem_class.__name__, identifier))
                # return a list instead
                if not unique:
                    return ListClientMixin(self.uri, self.elem_class, document)
                for elem in document.data[self.elem_class.path]:
                    if 'is_defining_ontology' in elem and elem['is_defining_ontology']:
                        return self.elem_class(**elem)
                return None
            return self.elem_class(**document.data)
        except ErrorMessage as e:
            # print(e.error)
            raise exceptions.OlsException(e.error)


class ListClientMixin(BaseClient):
    page_size = 100

    def __init__(self, uri, elem_class, document=None):
        super().__init__()
        self.document = document or self.client.get('/'.join([site, uri]), force_codec='hal')
        self.elem_class = elem_class
        self.index = 0

    def elem_class_instance(self, **data):
        return self.elem_class(**data)

    def __call__(self, filters=None, action=None):
        """
        Allow to search for a list of helpers, retrieve self, wich is now a iterator on the actual list of related
        helpers
        """
        if filters is None:
            filters = {}
        params = {'page': 0, 'size': self.page_size}
        try:
            check_fn = getattr(sys.modules[__name__], 'filters_' + self.path)
            if callable(check_fn):
                filters = check_fn(filters)
        except AttributeError:
            pass
        except AssertionError as e:
            raise exceptions.BadParameter({'error': "Bad Request", 'message': str(e),
                                           'status': 400, 'path': self.path, 'timestamp': time.time()})

        params.update(filters)
        path = action if action else self.path
        self.fetch_document(path, params)
        self.index = 0

        return self

    def fetch_document(self, path, params=None):
        if params is None:
            params = {}
        self.document = self.client.action(self.document, [path], params=params, validate=False)

    def fetch_page(self, page):
        self.document = self.client.get(
            '/'.join([site, self.path]) + '?page={}&size={}'.format(page, self.page_size),
            force_codec='hal')

    def __len__(self):
        return self.document['page']['totalElements']

    def __iter__(self):
        return self

    @property
    def path(self):
        return self.elem_class.path

    @property
    def page(self):
        return self.document['page']['number']

    @property
    def pages(self):
        return self.document['page']['totalPages'] - 1

    @property
    def data(self):
        if self.path in self.document:
            return self.document.data[self.path]
        return []

    def __next__(self):
        if self.index < len(self.data):
            # Simply return current indexed item
            pass
        elif self.page < self.pages:
            self.fetch_document('next')
            # self.document = self.client.action(self.document, ['next'])
            self.index = 0
        else:
            raise StopIteration
        loaded = self.elem_class_instance(**self.data[self.index])
        self.index += 1
        return loaded

    def __getitem__(self, item):
        if isinstance(item, slice):
            return [self[ii] for ii in range(*item.indices(len(self)))]
        if type(item) is not int:
            raise TypeError("Key indexes must be int, not {}".format(type(item)))
        if item > (len(self) - 1):
            # TODO relace with proper ArrayError Text
            raise KeyError("No corresponding key {}".format(item))
        else:
            page = item // self.page_size
            index = item % self.page_size
            if page == self.page:
                return self.elem_class_instance(**self.data[index])
            else:
                self.fetch_page(page)
                self.index = index
                return self.elem_class_instance(**self.data[index])

    def __repr__(self) -> str:
        return '[' + ','.join([repr(self.elem_class_instance(**data)) for data in self.data]) + ']'


class SearchClientMixin(ListClientMixin):
    base_search_uri = ''

    def __call__(self, filters=None, query=None):
        if query is None:
            raise exceptions.BadParameter({'error': "Bad Request", 'message': 'Missing query',
                                           'status': 400, 'path': 'search', 'timestamp': time.time()})
        self.query = query  #
        return super().__call__(filters)

    def __len__(self):
        return self.document[self.path]['numFound']

    def elem_class_instance(self, **kwargs):
        import ebi.ols.api.helpers as helpers
        type_item = kwargs.get('type')
        if type_item == 'property':
            return helpers.Property(**kwargs)
        elif type_item == 'individual':
            return helpers.Individual(**kwargs)
        elif type_item == 'ontology':
            return helpers.Ontology(**kwargs)
        else:
            return helpers.Term(**kwargs)

    @property
    def start(self):
        return self.document[self.path]['start']

    @property
    def path(self):
        return 'response'

    @property
    def page(self):
        return math.floor(self.start / self.page_size) + 1

    @property
    def pages(self):
        return math.ceil(len(self) / self.page_size)

    @property
    def data(self):
        return self.document.data[self.path]['docs']

    def fetch_document(self, path, params=None):
        if params is None:
            params = {}
        params.pop('page', 0)
        params.pop('size', 0)
        start = 0 if path != 'next' else self.start + self.page_size
        uri = '/'.join([site, 'search'])
        uri += '?q=' + self.query
        filters_uri = ''
        for filter_name, filter_value in params.items():
            # print(filter_name, filter_value)
            filters_uri += '&' + filter_name + '=' + urllib.parse.quote_plus(filter_value)
        filters_uri += "&exact=on"
        uri += filters_uri + '&rows=100&start={}'.format(start)
        self.base_search_uri = uri
        # print('Search uri', uri)
        self.document = self.client.get(uri, format='hal')

    def fetch_page(self, page):
        uri = self.base_search_uri + '&rows=100&start={}'.format(page * self.page_size)
        self.document = self.client.get(uri, format='hal')
